package cn.iocoder.yudao.module.system.service.robot;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.util.ObjectUtil;
import cn.iocoder.yudao.framework.common.util.date.DateUtils;
import cn.iocoder.yudao.framework.mybatis.core.query.LambdaQueryWrapperX;
import cn.iocoder.yudao.framework.mybatis.core.util.MyBatisUtils;
import cn.iocoder.yudao.module.system.controller.admin.robot.vo.*;
import cn.iocoder.yudao.module.system.dal.dataobject.houselocation.WareHouseLocationDO;
import cn.iocoder.yudao.module.system.dal.dataobject.robot.RobotInformationDO;
import cn.iocoder.yudao.module.system.dal.dataobject.robot.RobotTaskDO;
import cn.iocoder.yudao.module.system.dal.dataobject.robot.RobotTaskDetailDO;
import cn.iocoder.yudao.module.system.dal.mysql.houselocation.WareHouseLocationMapper;
import cn.iocoder.yudao.module.system.dal.mysql.robot.RobotInformationMapper;
import cn.iocoder.yudao.module.system.dal.mysql.robot.RobotTaskDetailMapper;
import cn.iocoder.yudao.module.system.dal.mysql.robot.RobotTaskMapper;
import cn.iocoder.yudao.module.system.enums.redis.RobotCacheLockEnum;
import cn.iocoder.yudao.module.system.enums.robot.*;
import cn.iocoder.yudao.module.system.util.redis.RedisUtil;
import cn.iocoder.yudao.module.system.util.redis.RedissonUtils;
import com.baomidou.mybatisplus.core.metadata.IPage;
import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.redisson.api.RLock;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;
import javax.annotation.Resource;

import org.springframework.transaction.annotation.Transactional;
import org.springframework.validation.annotation.Validated;

import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

import cn.iocoder.yudao.framework.common.pojo.PageResult;
import cn.iocoder.yudao.framework.common.util.object.BeanUtils;


import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception;
import static cn.iocoder.yudao.framework.common.exception.util.ServiceExceptionUtil.exception0;
import static cn.iocoder.yudao.module.system.enums.ErrorCodeConstants.*;

/**
 * 机器人任务主表 Service 实现类
 *
 * @author 陈宾顺
 */
@Slf4j
@Service
@Validated
public class RobotTaskServiceImpl implements RobotTaskService {

    private final String RESULT = "SUCCESS";

    @Resource
    private RobotTaskMapper taskMapper;

    @Resource
    private WareHouseLocationMapper locationMapper;

    @Resource
    private RedisUtil redisUtil;

    @Resource
    private RedissonUtils redissonUtils;

    @Resource
    private RobotTaskDetailMapper taskDetailMapper;

    @Resource
    private RobotTaskDetailService robotTaskDetailService;

    @Resource
    private RobotInformationMapper informationMapper;

    @Value("${zn.task-no:ZN}")
    private String taskNo;
    @Value("${zn.do_cycle:true}")
    private Boolean doCycle;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Long createTask(RobotTaskSaveReqVO createReqVO) throws InterruptedException {
        //校验
        doTaskCheck(createReqVO);

        //设置任务号
        if (ObjectUtil.isEmpty(createReqVO.getTaskNo())) {
            String incrementByKey = redisUtil.getIncrementByKey(RobotCacheLockEnum.TASK_NO.getKey());
            createReqVO.setTaskNo(taskNo+ DateUtils.getYearMonthDay()+incrementByKey);
        }
        //获取库位id
        RLock lock = redissonUtils.getLock(RobotCacheLockEnum.ROBOT_TASK_ADD_LOCK.getKey());

        String addResult = "";
        if (lock.tryLock(60l,TimeUnit.MINUTES)){
            try {
                addResult = addTask(createReqVO);
            } catch (Exception e) {
                log.error("下发任务出现异常 :{}",e);
                addResult = e.getMessage();
            } finally {
                lock.unlock();
            }
        }else {
            throw exception(REDISSON_NOT_OBTAIN_LOCK);
        }

        if (!RESULT.equals(addResult)) {
            throw exception0(TASK_CHECK_EXCEPTION.getCode(), addResult);
        }

        return 1l;
    }

    /**
     * 校验
     * @param createReqVO
     */
    private void doTaskCheck(RobotTaskSaveReqVO createReqVO) {
        //校验明细
        if (ObjectUtil.isEmpty(createReqVO.getTaskDetailList())) {
            throw exception(TASK_DATA_NOT_FULL);
        }
        //校验循环任务的配置未开启
        if (DoCycleEnum.YES.getType().equals(createReqVO.getDoCycle()) && !doCycle) {
            throw exception(TASK_CYCLE_NOT_OPEN);
        }
        //搬空任务只能选择线库或者区域
        if (MoveAllEnum.YES.getType().equals(createReqVO.getDoMoveAll())) {
            Set<Integer> releaseTypeList =
                    createReqVO.getTaskDetailList().stream().map(v -> v.getReleaseType())
                            .collect(Collectors.toSet());
            Set<Integer> takeTypeList =
                    createReqVO.getTaskDetailList().stream().map(v -> v.getTakeType())
                            .collect(Collectors.toSet());
            if (releaseTypeList.contains(ReleaseTakeEnum.TO_LOCATION.getType())
                    || takeTypeList.contains(ReleaseTakeEnum.TO_LOCATION.getType())) {
                throw exception(TASK_ONLY_CHOOSE_LOCATION);
            }
        }
        //校验机器人是否禁用
        Set<String> robotNos = createReqVO.getTaskDetailList()
                .stream()
                .map(RobotTaskDetailAddVO::getRobotNo)
                .collect(Collectors.toSet());
        if (ObjectUtil.isNotEmpty(robotNos)) {
            List<RobotInformationDO> robots = informationMapper.selectByRobotNos(robotNos);
            String robotNo = robots.stream()
                    .filter(v -> RobotTaskModelEnum.REJECTION.getType().equals(v.getRobotTaskModel()))
                    .map(RobotInformationDO::getRobotNo)
                    .collect(Collectors.joining(" "));

            if (ObjectUtil.isNotEmpty(robotNo)) {
                throw exception0(TASK_CHECK_EXCEPTION.getCode(), "以下机器人已被禁用 ",robotNo);
            }
        }
        //校验任务号是否重复
        if (ObjectUtil.isNotEmpty(createReqVO.getTaskNo())) {
            List<RobotTaskDO> taskDOS = taskMapper.selectList(new LambdaQueryWrapperX<RobotTaskDO>()
                    .eq(RobotTaskDO::getTaskNo, createReqVO.getTaskNo()));
            if (ObjectUtil.isNotEmpty(taskDOS)) {
                throw exception(TASK_CHECK_EXIST_NO);
            }
        }
    }

    /**
     * 添加任务
     * @param createReqVO
     */
    @Transactional(rollbackFor = Exception.class)
    public String addTask(RobotTaskSaveReqVO createReqVO) {
        RobotTaskDO task = BeanUtils.toBean(createReqVO, RobotTaskDO.class);
        taskMapper.insert(task);

        // todo 更新数据库顺序 后续改为xxl-job一致

        //查找库位
        List<Long> locationIds = new ArrayList<>();
        if (MoveAllEnum.NO.getType().equals(createReqVO.getDoMoveAll())) {
            locationIds = setSingleLocationIdNo(createReqVO.getTaskDetailList(),task);
        } else {
            locationIds = setAllLocationIdNo(createReqVO,task.getId());
        }

        // 插入
        createReqVO.getTaskDetailList()
                .forEach(taskDetail -> {
                    taskDetail.setRobotTaskId(task.getId());
                });
        robotTaskDetailService.insertRobotDetailList(createReqVO.getTaskDetailList());

        List<List<Long>> fromPartition = Lists.partition(locationIds, 100);
        fromPartition.forEach(list-> {
            locationMapper.updateLocationLockList(list,task.getId(),LocationLockEnum.NO.getType());
        });

        return RESULT;
    }

    /**
     * 设置放货和取货的id
     * @param createReqVO
     * @param taskId
     */
    @Transactional(rollbackFor = Exception.class)
    public List<Long> setAllLocationIdNo(RobotTaskSaveReqVO createReqVO, Long taskId) {
        List<RobotTaskDetailAddVO> taskDetailList = createReqVO.getTaskDetailList();

        List<RobotTaskDetailAddVO> newTaskDetailList = new ArrayList<>();
        List<Long> toLocationIds = new ArrayList<>(); //被此次任务锁定的空库位id
        List<Long> fromLocationIds = new ArrayList<>(); //被此次任务锁定的有货库位id
        for (RobotTaskDetailAddVO robotTaskVo : taskDetailList) {
            Set<Long> mapIds = new HashSet<>();
            if (ObjectUtil.isNotEmpty(robotTaskVo.getRobotNo())) {
                RobotInformationPageReqVO pageReqVO = new RobotInformationPageReqVO();
                pageReqVO.setRobotNo(robotTaskVo.getRobotNo());
                mapIds = informationMapper.selectPage(pageReqVO).getList()
                        .stream()
                        .findFirst()
                        .map(RobotInformationDO::getFloorAreaJson)
                        .orElse(new HashSet<>());
            }
            //查找有货且非锁定的库位
            WareHouseLocationDO query = null;
            if (ReleaseTakeEnum.TO_LANE.getType().equals(robotTaskVo.getTakeType())) {
                query = WareHouseLocationDO.builder()
                        .laneId(robotTaskVo.getTakeId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.YES.getType())
                        .build();
            } else if (ReleaseTakeEnum.TO_AREA.getType().equals(robotTaskVo.getTakeType())) {
                query = WareHouseLocationDO.builder()
                        .areaId(robotTaskVo.getTakeId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.YES.getType())
                        .build();
            }

            List<WareHouseLocationDO> stockList = locationMapper.selectLocations(query,toLocationIds,mapIds);
            if (ObjectUtil.isEmpty(stockList)) {
                log.error("取货线库/区域为空或者已锁定或者机器人取放货区域受限制 :{}",robotTaskVo.toString());
                throw new RuntimeException("取货线库/区域为空或者已锁定或者机器人取放货区域受限制");
            }
            //判断取货库位是否存在未完成的任务
            Set<Long> stockLocationIds = stockList.stream().map(WareHouseLocationDO::getId).collect(Collectors.toSet());
            List<RobotTaskDetailDO> doingTasks = taskDetailMapper.queryDoingTaskByIds(stockLocationIds);
            if (ObjectUtil.isNotEmpty(doingTasks)) {
                throw new RuntimeException("取货线库/区域,存在未完成的任务");
            }
            //查找空库位
            WareHouseLocationDO releaseQuery = null;
            if (ReleaseTakeEnum.TO_LANE.getType().equals(robotTaskVo.getReleaseType())) {
                releaseQuery = WareHouseLocationDO.builder()
                        .laneId(robotTaskVo.getReleaseId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.NO.getType())
                        .build();
            } else if (ReleaseTakeEnum.TO_AREA.getType().equals(robotTaskVo.getReleaseType())) {
                 releaseQuery = WareHouseLocationDO.builder()
                        .areaId(robotTaskVo.getReleaseId())
                        .locationEnable(LocationEnableEnum.YES.getType())
                        .locationLock(LocationLockEnum.YES.getType())
                        .locationUseStatus(LocationUseStatusEnum.NO.getType())
                        .build();
            }

            List<WareHouseLocationDO> releaseStockList = locationMapper.selectLocations(releaseQuery,toLocationIds,mapIds);
            if (ObjectUtil.isEmpty(releaseStockList) || releaseStockList.size() < stockList.size()) {
                log.error("放货线库/区域库位数量不足或者机器人取放货区域受限制 :{}",robotTaskVo.toString());
                throw new RuntimeException("放货线库/区域库位数量不足或者机器人取放货区域受限制");
            }
            //赋值数据
            List<RobotTaskDetailAddVO> addTaskDetailList = getTaskDatail(stockList,releaseStockList,toLocationIds,
                                                                            robotTaskVo,taskId,fromLocationIds);
            newTaskDetailList.addAll(addTaskDetailList);

        }

        createReqVO.setTaskDetailList(newTaskDetailList);

        if (ObjectUtil.isNotEmpty(fromLocationIds)) {
            toLocationIds.addAll(fromLocationIds);
        }
        return toLocationIds;
    }

    /**
     * 赋值数据
     * @param stockList
     * @param releaseStockList
     * @param toLocationIds
     * @param robotTaskVo
     * @return
     */
    private List<RobotTaskDetailAddVO> getTaskDatail(List<WareHouseLocationDO> stockList,
                                                     List<WareHouseLocationDO> releaseStockList,
                                                     List<Long> toLocationIds, RobotTaskDetailAddVO robotTaskVo,
                                                     Long taskId, List<Long> fromLocationIds) {
        List<RobotTaskDetailAddVO> addTaskDetailList = new ArrayList<>();
        for (int i = 0; i < stockList.size(); i++) {
            RobotTaskDetailAddVO robotTaskDetailAddVo = new RobotTaskDetailAddVO();
            robotTaskDetailAddVo.setTaskType(robotTaskVo.getTaskType());
            robotTaskDetailAddVo.setReleaseType(robotTaskVo.getReleaseType());
            robotTaskDetailAddVo.setTakeType(robotTaskVo.getTakeType());
            robotTaskDetailAddVo.setTakeId(robotTaskVo.getTakeId());
            robotTaskDetailAddVo.setReleaseId(robotTaskVo.getReleaseId());
            robotTaskDetailAddVo.setRobotNo(robotTaskVo.getRobotNo());
            robotTaskDetailAddVo.setFromLocationNo(stockList.get(i).getLocationNo());
            robotTaskDetailAddVo.setFromLocationId(stockList.get(i).getId());
            robotTaskDetailAddVo.setToLocationNo(releaseStockList.get(i).getLocationNo());
            robotTaskDetailAddVo.setToLocationId(releaseStockList.get(i).getId());
            robotTaskDetailAddVo.setFromLaneId(stockList.get(i).getLaneId());
            robotTaskDetailAddVo.setToLaneId(releaseStockList.get(i).getLaneId());
            robotTaskDetailAddVo.setLocationNumber(stockList.get(i).getLocationNumber());
            robotTaskDetailAddVo.setRobotTaskId(taskId);
            robotTaskDetailAddVo.setPriority(robotTaskVo.getPriority());
            robotTaskDetailAddVo.setFromLocationStorey(stockList.get(i).getLocationStorey());
            robotTaskDetailAddVo.setToLocationStorey(releaseStockList.get(i).getLocationStorey());
            addTaskDetailList.add(robotTaskDetailAddVo);

            toLocationIds.add(releaseStockList.get(i).getId());
            fromLocationIds.add(stockList.get(i).getId());
        }
        return addTaskDetailList;
    }

    /**
     * 设置放货和取货的id
     * @param taskDetailList
     * @return
     */
    @Transactional(rollbackFor = Exception.class)
    public List<Long> setSingleLocationIdNo(List<RobotTaskDetailAddVO> taskDetailList, RobotTaskDO task){
        List<Long> locationIds = new ArrayList<>();
        for (RobotTaskDetailAddVO robotTaskVo : taskDetailList) {
            robotTaskVo.setRobotTaskId(task.getId());
            robotTaskVo.setPriority(task.getPriority());
            RobotTaskTypeEnum robotTaskType = RobotTaskTypeEnum.getRobotTaskType(robotTaskVo.getTaskType());
            switch (robotTaskType){
                case TAKE_RELEASE:
                    doTakeRelease(robotTaskVo,locationIds);
                    break;
                case PARK:
                    doPark(robotTaskVo,locationIds);
                    break;
                case CHARGE:
                    doCharge(robotTaskVo);
                    break;
                case MOVE:
                    doMove(robotTaskVo,locationIds);
                    break;
                case TAKE:
                    doTake(robotTaskVo,locationIds);
                    break;
                case RELEASE:
                    doRelease(robotTaskVo,locationIds);
                    break;
                case SCAN:
                    doScan(robotTaskVo,locationIds);
                    break;
                case DETECTING_TRAYS:
                    doDetectingTrays(robotTaskVo,locationIds);
                    break;
                default :
                    log.error("任务类型不存在 ");
                    throw new RuntimeException("任务类型不存在");
            }
        }
        return locationIds;
    }

    /**
     * 检测托盘类型
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doDetectingTrays(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        setToLocation(robotTaskVo,getMapIdsByRobotNo(robotTaskVo.getRobotNo()),new ArrayList<>());
        locationIds.add(robotTaskVo.getToLocationId());
    }

    /**
     * 扫描码
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doScan(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        setToLocation(robotTaskVo,getMapIdsByRobotNo(robotTaskVo.getRobotNo()),new ArrayList<>());
        locationIds.add(robotTaskVo.getToLocationId());
    }

    /**
     * 仅放货
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doRelease(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        setToLocation(robotTaskVo,getMapIdsByRobotNo(robotTaskVo.getRobotNo()),locationIds);
        locationIds.add(robotTaskVo.getToLocationId());
    }

    /**
     * 仅取货
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doTake(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        Set<Long> mapIds = new HashSet<>();
        if (ObjectUtil.isNotEmpty(robotTaskVo.getRobotNo())) {
            mapIds = getMapIdsByRobotNo(robotTaskVo.getRobotNo());
        }
        setFromLocation(robotTaskVo,mapIds,locationIds);
        locationIds.add(robotTaskVo.getFromLocationId());
    }

    /**
     * 移动
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doMove(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        setToLocation(robotTaskVo,getMapIdsByRobotNo(robotTaskVo.getRobotNo()),locationIds);
        locationIds.add(robotTaskVo.getToLocationId());
    }

    /**
     * 充电
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doCharge(RobotTaskDetailAddVO robotTaskVo) {

    }

    /**
     * 停车
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doPark(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        setToLocation(robotTaskVo,getMapIdsByRobotNo(robotTaskVo.getRobotNo()),locationIds);
        locationIds.add(robotTaskVo.getToLocationId());
    }

    /**
     * 查询此机器人对应的楼层/区域(json)
     * @param robotNo
     * @return
     */
    public Set<Long> getMapIdsByRobotNo(String robotNo) {
        if (ObjectUtil.isNotEmpty(robotNo)) {
            RobotInformationPageReqVO pageReqVO = new RobotInformationPageReqVO();
            pageReqVO.setRobotNo(robotNo);
            return  informationMapper.selectPage(pageReqVO).getList()
                    .stream()
                    .findFirst()
                    .map(RobotInformationDO::getFloorAreaJson)
                    .orElse(new HashSet<>());
        }
        return new HashSet<>();
    }

    /**
     * 设置停车/目标库位/扫描库位
     */
    @Transactional(rollbackFor = Exception.class)
    public void setToLocation(RobotTaskDetailAddVO robotTaskVo, Set<Long> mapIds, List<Long> locationIds){
        if (ReleaseTakeEnum.TO_LOCATION.getType().equals(robotTaskVo.getReleaseType())) {
            WareHouseLocationDO query = WareHouseLocationDO.builder().id(robotTaskVo.getReleaseId()).build();
            WareHouseLocationDO locationDO = locationMapper.queryAllByLimit(query);
            robotTaskVo.setToLocationNo(locationDO.getLocationNo());
            robotTaskVo.setToLocationId(robotTaskVo.getReleaseId());
            robotTaskVo.setToLocationStorey(locationDO.getLocationStorey());
            robotTaskVo.setToLaneId(locationDO.getLaneId());
            if (ObjectUtil.isNotEmpty(mapIds) && !mapIds.contains(locationDO.getMapId())) {
                log.error("机器人不能在此放货库位放货 :{}, 机器人编号 :{}", locationDO.getLocationNo(),robotTaskVo.getRobotNo());
                throw new RuntimeException("机器人不能在此放货库位放货 "+ robotTaskVo.getReleaseId());
            }
        } else {
            WareHouseLocationDO wareHouseLocationDO =
                    locationMapper.selectByTypeAndId(LocationUseStatusEnum.NO.getType(), robotTaskVo.getReleaseType()
                            ,robotTaskVo.getReleaseId(),locationIds,mapIds);
            if (ObjectUtil.isEmpty(wareHouseLocationDO)) {
                log.error("放货库位为空或者已锁定或者机器人取放货区域受限制 :{}", robotTaskVo.getReleaseId());
                throw new RuntimeException("放货库位为空或者已锁定或者机器人取放货区域受限制 "+ robotTaskVo.getReleaseId());
            }
            robotTaskVo.setToLocationNo(wareHouseLocationDO.getLocationNo());
            robotTaskVo.setToLocationId(wareHouseLocationDO.getId());
            robotTaskVo.setToLocationStorey(wareHouseLocationDO.getLocationStorey());
            robotTaskVo.setToLaneId(wareHouseLocationDO.getLaneId());
        }
    }

    /**
     * 设置取货库位
     * @param robotTaskVo
     * @param mapIds
     * @param locationIds
     */
    @Transactional(rollbackFor = Exception.class)
    public void setFromLocation(RobotTaskDetailAddVO robotTaskVo, Set<Long> mapIds, List<Long> locationIds) {
        if (ReleaseTakeEnum.TO_LOCATION.getType().equals(robotTaskVo.getTakeType())) {
            WareHouseLocationDO query = WareHouseLocationDO.builder().id(robotTaskVo.getTakeId()).build();
            WareHouseLocationDO locationDO = locationMapper.queryAllByLimit(query);
            robotTaskVo.setFromLocationNo(locationDO.getLocationNo());
            robotTaskVo.setFromLocationId(robotTaskVo.getTakeId());
            robotTaskVo.setFromLocationStorey(locationDO.getLocationStorey());
            robotTaskVo.setFromLaneId(locationDO.getLaneId());
            robotTaskVo.setLocationNumber(locationDO.getLocationNumber());
            if (ObjectUtil.isNotEmpty(mapIds) && !mapIds.contains(locationDO.getMapId())) {
                log.error("机器人不能在此取货库位取货 :{}, 机器人编号 :{}", locationDO.getLocationNo(),robotTaskVo.getRobotNo());
                throw new RuntimeException("机器人不能在此取货库位取货 "+ robotTaskVo.getTakeId());
            }
        } else {
            WareHouseLocationDO wareHouseLocationDO =
                    locationMapper.selectByTypeAndId(LocationUseStatusEnum.YES.getType(),robotTaskVo.getTakeType(),
                            robotTaskVo.getTakeId(),locationIds,mapIds);
            if (ObjectUtil.isEmpty(wareHouseLocationDO)) {
                log.error("取货库位为空或者已锁定或者机器人取放货区域受限制 :{}", robotTaskVo.getTakeId());
                throw new RuntimeException("取货库位为空或者已锁定或者机器人取放货区域受限制 "+ robotTaskVo.getTakeId());
            }
            robotTaskVo.setFromLocationNo(wareHouseLocationDO.getLocationNo());
            robotTaskVo.setFromLocationId(wareHouseLocationDO.getId());
            robotTaskVo.setFromLocationStorey(wareHouseLocationDO.getLocationStorey());
            robotTaskVo.setFromLaneId(wareHouseLocationDO.getLaneId());
            robotTaskVo.setLocationNumber(wareHouseLocationDO.getLocationNumber());
        }
    }

    /**
     * 取放
     * @param robotTaskVo
     */
    @Transactional(rollbackFor = Exception.class)
    public void doTakeRelease(RobotTaskDetailAddVO robotTaskVo, List<Long> locationIds) {
        Set<Long> mapIds = new HashSet<>();
        if (ObjectUtil.isNotEmpty(robotTaskVo.getRobotNo())) {
            mapIds = getMapIdsByRobotNo(robotTaskVo.getRobotNo());
        }

        //校验放货库位
        setToLocation(robotTaskVo, mapIds,locationIds);
        locationIds.add(robotTaskVo.getToLocationId());

        //校验取货库位
        setFromLocation(robotTaskVo,mapIds,locationIds);
        locationIds.add(robotTaskVo.getFromLocationId());

        //验证取货库位是否存在未完成的任务
        Set<Long> set  = new HashSet<>();
        set.add(robotTaskVo.getFromLocationId());
        List<RobotTaskDetailDO> doingTasks = taskDetailMapper.queryDoingTaskByIds(set);
        if (ObjectUtil.isNotEmpty(doingTasks)) {
            log.error("取货库位,存在未完成的任务 :{}", robotTaskVo.getTakeId());
            throw new RuntimeException("取货库位,存在未完成的任务 ");
        }

    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void updateTask(RobotTaskSaveReqVO updateReqVO) {
        RLock lock = redissonUtils.getLock(RobotCacheLockEnum.ROBOT_TASK_DISTRIBUTE_LOCK.getKey());
        String msg = "";
        if (lock.tryLock()){
            try {
                // 校验存在
                validateTaskExists(updateReqVO.getId());
                if (ObjectUtil.isEmpty(updateReqVO.getId())) {
                    throw exception(TASK_CHECK_ID_EXCEPTION);
                }
                // 更新
                //修改优先级需要订单状态为未开始/处理中
                RobotTaskDO robotTaskDO = taskMapper.selectById(updateReqVO.getId());
                List<RobotTaskDetailDO> taskDetailDOS = taskDetailMapper.queryByTaskId(updateReqVO.getId());
                if (ObjectUtil.isNotEmpty(updateReqVO.getPriority())
                        && !updateReqVO.getPriority().equals(robotTaskDO.getPriority())) {
                    if (!RobotTaskStatusEnum.NEW.getType().equals(robotTaskDO.getTaskStatus())
                        && !RobotTaskStatusEnum.DOING.getType().equals(robotTaskDO.getTaskStatus())) {
                        throw exception(TASK_CHECK_TASK_PRIORITY);
                    }
                    for (RobotTaskDetailDO taskDetailDO : taskDetailDOS) {
                        taskDetailDO.setPriority(updateReqVO.getPriority());
                    }
                }
                //修改订单状态
                if (ObjectUtil.isNotEmpty(updateReqVO.getTaskStatus())
                        && RobotTaskStatusEnum.DONE.getType().equals(robotTaskDO.getTaskStatus())
                        && !robotTaskDO.getTaskStatus().equals(updateReqVO.getTaskStatus())) {
                    throw exception(TASK_CHECK_TASK_STATUS);
                }

                List<Long> locationIds = new ArrayList<>();
                Set<String> robotNos = new HashSet<>();
                if (ObjectUtil.isNotEmpty(updateReqVO.getTaskStatus())
                    && RobotTaskStatusEnum.CLOSE.getType().equals(updateReqVO.getTaskStatus())
                    && !robotTaskDO.getTaskStatus().equals(updateReqVO.getTaskStatus())) {
                    for (RobotTaskDetailDO taskDetailDO : taskDetailDOS) {
                        if (RobotTaskDetailStatusEnum.DONE.getType().equals(taskDetailDO.getTaskStatus())) {
                            continue;
                        }
                        //后期需做状态映射
                        taskDetailDO.setTaskStatus(updateReqVO.getTaskStatus());

                        // todo 关闭需要下发任务给车机   。 还需要考虑下处理中的任务怎么处理

                        if (ObjectUtil.isNotEmpty(taskDetailDO.getFromLocationId())) {
                            locationIds.add(taskDetailDO.getFromLocationId());
                        }
                        if (ObjectUtil.isNotEmpty(taskDetailDO.getToLocationId())) {
                            locationIds.add(taskDetailDO.getToLocationId());
                        }
                        if (ObjectUtil.isNotEmpty(taskDetailDO.getRobotNo())) {
                            robotNos.add(taskDetailDO.getRobotNo());
                        }
                    }
                }

                RobotTaskDO updateObj = BeanUtils.toBean(updateReqVO, RobotTaskDO.class);
                taskMapper.updateById(updateObj);

                taskDetailMapper.updateBatch(taskDetailDOS);

                if (ObjectUtil.isNotEmpty(locationIds)) {
                    locationMapper.updateLocationLockList(locationIds,robotTaskDO.getId()
                        ,LocationLockEnum.YES.getType());
                }

                // todo 执行中的需要锁车。 需要添加执行中的判断
                if (ObjectUtil.isNotEmpty(robotNos)) {
//                    informationMapper.updateRobotTaskModel(robotNos,RobotTaskModelEnum.REJECTION.getType());
                }

            } catch (Exception e) {
                log.error("更新任务失败 :{}",e.getMessage());
                msg = e.getMessage();
            } finally {
                lock.unlock();
            }
        } else {
            throw exception(REDISSON_NOT_OBTAIN_LOCK);
        }

        if (ObjectUtil.isNotEmpty(msg)) {
            throw exception0(TASK_CHECK_UPDATE_STATUS.getCode(), msg);
        }
    }

    @Override
    public void deleteTask(Long id) {
        // 校验存在
        validateTaskExists(id);
        // 删除
        taskMapper.deleteById(id);
    }

    private void validateTaskExists(Long id) {
        if (taskMapper.selectById(id) == null) {
            throw exception(TASK_NOT_EXISTS);
        }
    }

    @Override
    public RobotTaskDO getTask(Long id) {
        return taskMapper.selectById(id);
    }

    @Override
    public PageResult<RobotTaskRespVO> getTaskPage(RobotTaskPageReqVO pageReqVO) {

        PageResult<RobotTaskDO> pageResult = null;
        PageResult<RobotTaskRespVO> dataPage = new PageResult<>();
        List<RobotTaskDO> list = new ArrayList<>();
        if (ObjectUtil.isEmpty(pageReqVO.getRobotNo())) {
            pageResult = taskMapper.selectPage(pageReqVO);
            dataPage.setTotal(pageResult.getTotal());
            list = pageResult.getList();
        }else {
            IPage mpPage = MyBatisUtils.buildPage(pageReqVO);
            IPage<RobotTaskDO> page = taskMapper.selectPageList(mpPage, pageReqVO);
            dataPage.setTotal(page.getTotal());
            list = page.getRecords();
        }

        List<RobotTaskRespVO> targetList = BeanUtil.copyToList(list, RobotTaskRespVO.class);
        for (RobotTaskRespVO robotTaskRespVO : targetList) {
            robotTaskRespVO.setDetails(taskDetailMapper.queryByTaskId(robotTaskRespVO.getId()));
        }
        dataPage.setList(targetList);
        return dataPage;
    }

    @Override
    public String getTaskNo() {
        String incrementByKey = redisUtil.getIncrementByKey(RobotCacheLockEnum.TASK_NO.getKey());
        return taskNo+ DateUtils.getYearMonthDay()+incrementByKey;
    }

}